@@ -20,6 +20,7 @@ import apt
 from UbuntuDrivers import kerneldetection
 
 system_architecture = apt.apt_pkg.get_architectures()[0]
+lookup_cache = {}
 
 
 def system_modaliases(sys_path=None):
@@ -48,10 +49,10 @@ def system_modaliases(sys_path=None):
         # devices on SSB bus only mention the modalias in the uevent file (as
         # of 2.6.24)
         elif 'ssb' in path and 'uevent' in files:
-            with open(os.path.join(path, 'uevent')) as f:
-                for l in f:
-                    if l.startswith('MODALIAS='):
-                        modalias = l.split('=', 1)[1].strip()
+            with open(os.path.join(path, 'uevent')) as fd:
+                for line in fd:
+                    if line.startswith('MODALIAS='):
+                        modalias = line.split('=', 1)[1].strip()
                         break
 
         if not modalias:
@@ -171,6 +172,10 @@ def _is_package_free(pkg):
     # it would be better to check the actual license, as we do not have
     # the component for third-party packages; but this is the best we can do
     # at the moment
+    #
+    # We can assume the NVIDIA packages to be non-free
+    if pkg.name.startswith('nvidia'):
+        return False
     for o in pkg.candidate.origins:
         if o.component in ('restricted', 'multiverse'):
             return False
@@ -205,6 +210,38 @@ def _pkg_get_module(pkg):
     return module
 
 
+def _pkg_get_support(pkg):
+    '''Determine support level from apt Package object'''
+
+    try:
+        support = pkg.candidate.record['Support']
+    except (KeyError, AttributeError):
+        logging.debug('_pkg_get_support %s: package has no Support header, cannot determine support level', pkg.name)
+        return None
+
+    if support not in ('NFB', 'LTSB'):
+        logging.warning('_pkg_get_support %s: package has invalid Support %s'
+                        'header, cannot determine support level', pkg.name, support)
+        return None
+
+    return support
+
+
+def _is_runtimepm_supported(pkg, alias):
+    '''Check if the package supports runtimepm for the given modalias'''
+    try:
+        m = pkg.candidate.record['PmAliases']
+    except (KeyError, AttributeError, UnicodeDecodeError):
+        return False
+    else:
+        if m.find('nvidia(') != 0:
+            return False
+
+        n = m[m.find('(')+1: m.find(')')]
+        modaliases = n.split(', ')
+        return any(fnmatch.fnmatch(alias.lower(), regex.lower()) for regex in modaliases)
+
+
 def _is_manual_install(pkg):
     '''Determine if the kernel module from an apt.Package is manually installed.'''
 
@@ -222,13 +259,18 @@ def _is_manual_install(pkg):
     if not module:
         return False
 
-    modinfo = subprocess.Popen(['modinfo', module], stdout=subprocess.PIPE,
-                               stderr=subprocess.PIPE)
-    modinfo.communicate()
-    if modinfo.returncode == 0:
-        logging.debug('_is_manual_install %s: builds module %s which is available, manual install',
-                      pkg.name, module)
-        return True
+    try:
+        modinfo = subprocess.Popen(['modinfo', module], stdout=subprocess.PIPE,
+                                   stderr=subprocess.PIPE)
+        modinfo.communicate()
+    except (OSError, FileNotFoundError, subprocess.CalledProcessError) as e:
+        logging.debug('_is_manual_install failed: %s', str(e))
+        return False
+    else:
+        if modinfo.returncode == 0:
+            logging.debug('_is_manual_install %s: builds module %s which is available, manual install',
+                          pkg.name, module)
+            return True
 
     logging.debug('_is_manual_install %s: builds module %s which is not available, no manual install',
                   pkg.name, module)
@@ -299,10 +341,15 @@ def system_driver_packages(apt_cache=None, sys_path=None, freeonly=False, includ
                      versions; these have this flag, where exactly one has
                      recommended == True, and all others False.
     '''
+    global lookup_cache
     modaliases = system_modaliases(sys_path)
 
     if not apt_cache:
-        apt_cache = apt.Cache()
+        try:
+            apt_cache = apt.Cache()
+        except Exception as ex:
+            logging.error(ex)
+            return {}
 
     packages = {}
     for alias, syspath in modaliases.items():
@@ -316,6 +363,8 @@ def system_driver_packages(apt_cache=None, sys_path=None, freeonly=False, includ
                     'syspath': syspath,
                     'free': _is_package_free(p),
                     'from_distro': _is_package_from_distro(p),
+                    'support': _pkg_get_support(p),
+                    'runtimepm': _is_runtimepm_supported(p, alias)
                 }
             (vendor, model) = _get_db_name(syspath, alias)
             if vendor is not None:
@@ -326,19 +375,16 @@ def system_driver_packages(apt_cache=None, sys_path=None, freeonly=False, includ
     # Add "recommended" flags for NVidia alternatives
     nvidia_packages = [p for p in packages if p.startswith('nvidia-')]
     if nvidia_packages:
+        # Create a cache for looking up drivers to pick the best
+        # candidate
+        for key, value in packages.items():
+            if key.startswith('nvidia-'):
+                lookup_cache[key] = value
         nvidia_packages.sort(key=functools.cmp_to_key(_cmp_gfx_alternatives))
         recommended = nvidia_packages[-1]
         for p in nvidia_packages:
             packages[p]['recommended'] = (p == recommended)
 
-    # Add "recommended" flags for fglrx alternatives
-    fglrx_packages = [p for p in packages if p.startswith('fglrx-')]
-    if fglrx_packages:
-        fglrx_packages.sort(key=functools.cmp_to_key(_cmp_gfx_alternatives))
-        recommended = fglrx_packages[-1]
-        for p in fglrx_packages:
-            packages[p]['recommended'] = (p == recommended)
-
     # add available packages which need custom detection code
     for plugin, pkgs in detect_plugin_packages(apt_cache).items():
         for p in pkgs:
@@ -371,15 +417,22 @@ def _get_headless_no_dkms_metapackage(pkg, apt_cache):
     This is useful when dealing with packages such as nvidia-driver-$flavour
     whose headless-no-dkms metapackage would be nvidia-headless-no-dkms-$flavour
     '''
-    headless_template = 'nvidia-headless-no-dkms-%s'
     name = pkg.shortname
-    flavour = name[name.rfind('-')+1:]
-    try:
-        int(flavour)
-    except ValueError:
+
+    pattern = re.compile('nvidia-([0-9]+)')
+    match = pattern.match(name)
+    if match:
+        logging.debug('Legacy driver detected: %s. Skipping.', name)
         return metapackage
 
-    candidate = headless_template % flavour
+    pattern = re.compile('nvidia-driver-([0-9]+)(.*)')
+    match = pattern.match(name)
+    if not match:
+        logging.debug('No flavour can be found in %s. Skipping.', name)
+        return metapackage
+
+    candidate_flavour = '%s%s' % (match.group(1), match.group(2))
+    candidate = 'nvidia-headless-no-dkms-%s' % (candidate_flavour)
 
     try:
         package = apt_cache.__getitem__(candidate)
@@ -432,7 +485,11 @@ def system_device_specific_metapackages(apt_cache=None, sys_path=None, include_o
     modaliases = system_modaliases(sys_path)
 
     if not apt_cache:
-        apt_cache = apt.Cache()
+        try:
+            apt_cache = apt.Cache()
+        except Exception as ex:
+            logging.error(ex)
+            return {}
 
     packages = {}
     for alias, syspath in modaliases.items():
@@ -445,6 +502,7 @@ def system_device_specific_metapackages(apt_cache=None, sys_path=None, include_o
                     'free': _is_package_free(p),
                     'from_distro': _is_package_from_distro(p),
                     'recommended': True,
+                    'support': _pkg_get_support(p),
                 }
 
     return packages
@@ -483,11 +541,16 @@ def system_gpgpu_driver_packages(apt_cache=None, sys_path=None):
                      versions; these have this flag, where exactly one has
                      recommended == True, and all others False.
     '''
+    global lookup_cache
     vendors_whitelist = ['10de']
     modaliases = system_modaliases(sys_path)
 
     if not apt_cache:
-        apt_cache = apt.Cache()
+        try:
+            apt_cache = apt.Cache()
+        except Exception as ex:
+            logging.error(ex)
+            return {}
 
     packages = {}
     for alias, syspath in modaliases.items():
@@ -500,6 +563,7 @@ def system_gpgpu_driver_packages(apt_cache=None, sys_path=None):
                         'syspath': syspath,
                         'free': _is_package_free(p),
                         'from_distro': _is_package_from_distro(p),
+                        'support': _pkg_get_support(p),
                     }
                 if vendor is not None:
                     packages[p.name]['vendor'] = vendor
@@ -513,7 +577,12 @@ def system_gpgpu_driver_packages(apt_cache=None, sys_path=None):
     # Add "recommended" flags for NVidia alternatives
     nvidia_packages = [p for p in packages if p.startswith('nvidia-')]
     if nvidia_packages:
-        nvidia_packages.sort(key=functools.cmp_to_key(_cmp_gfx_alternatives))
+        # Create a cache for looking up drivers to pick the best
+        # candidate
+        for key, value in packages.items():
+            if key.startswith('nvidia-'):
+                lookup_cache[key] = value
+        nvidia_packages.sort(key=functools.cmp_to_key(_cmp_gfx_alternatives_gpgpu))
         recommended = nvidia_packages[-1]
         for p in nvidia_packages:
             packages[p]['recommended'] = (p == recommended)
@@ -576,7 +645,11 @@ def system_device_drivers(apt_cache=None, sys_path=None, freeonly=False):
     '''
     result = {}
     if not apt_cache:
-        apt_cache = apt.Cache()
+        try:
+            apt_cache = apt.Cache()
+        except Exception as ex:
+            logging.error(ex)
+            return {}
 
     # copy the system_driver_packages() structure into the by-device structure
     for pkg, pkginfo in system_driver_packages(apt_cache, sys_path,
@@ -609,29 +682,6 @@ def system_device_drivers(apt_cache=None, sys_path=None, freeonly=False):
     return result
 
 
-def auto_install_filter(packages):
-    '''Get packages which are appropriate for automatic installation.
-
-    Return the subset of the given list of packages which are appropriate for
-    automatic installation by the installer. This applies to e. g. the Broadcom
-    Wifi driver (as there is no alternative), but not to the FGLRX proprietary
-    graphics driver (as the free driver works well and FGLRX does not provide
-    KMS).
-    '''
-    # any package which matches any of those globs will be accepted
-    whitelist = ['bcmwl*', 'pvr-omap*', 'virtualbox-guest*', 'nvidia-*',
-                 'open-vm-tools*', 'oem-*-meta']
-    allow = []
-    for pattern in whitelist:
-        allow.extend(fnmatch.filter(packages, pattern))
-
-    result = {}
-    for p in allow:
-        if 'recommended' not in packages[p] or packages[p]['recommended']:
-            result[p] = packages[p]
-    return result
-
-
 class _GpgpuDriver(object):
 
     def __init__(self, vendor=None, flavour=None):
@@ -650,26 +700,22 @@ class _GpgpuDriver(object):
 def _process_driver_string(string):
     '''Returns a _GpgpuDriver object'''
     driver = _GpgpuDriver()
-    if string.find(':') != -1:
-        details = string.split(':')
-        # Remove empty strings
-        details = [x for x in details if x.strip()]
-        if len(details) != 2:
-            return None
-        for elem in details:
-            try:
-                int(elem)
-            except ValueError:
-                driver.vendor = elem
-            else:
-                driver.flavour = elem
-    else:
-        try:
-            int(string)
-        except ValueError:
-            driver.vendor = string
-        else:
-            driver.flavour = string
+
+    full_pattern = re.compile('(.+):([0-9]+)(.*)')
+    vendor_only_pattern = re.compile('([a-z]+)')
+    series_only_pattern = re.compile('([0-9]+)(.*)')
+
+    full_match = full_pattern.match(string)
+    vendor_match = vendor_only_pattern.match(string)
+    series_match = series_only_pattern.match(string)
+
+    if full_match:
+        driver.vendor = full_match.group(1)
+        driver.flavour = '%s%s' % (full_match.group(2), full_match.group(3))
+    elif vendor_match:
+        driver.vendor = string
+    elif series_match:
+        driver.flavour = '%s%s' % (series_match.group(1), series_match.group(2))
 
     return driver
 
@@ -697,15 +743,10 @@ def gpgpu_install_filter(packages, drivers_str):
     ubuntu-drivers autoinstall --gpgpu nvidia:390,amdgpu
     ubuntu-drivers autoinstall --gpgpu amdgpu:version
     '''
-
     if not packages:
         return result
 
-    # No args, just --gpgpu
-    if drivers_str == 'default':
-        driver = _GpgpuDriver()
-        drivers.append(driver)
-    else:
+    if drivers_str:
         # Just one driver
         # e.g. --gpgpu 390
         #      --gpgpu nvidia:390
@@ -716,6 +757,10 @@ def gpgpu_install_filter(packages, drivers_str):
             driver = _process_driver_string(item)
             if driver and driver.is_valid():
                 drivers.append(driver)
+    else:
+        # No args, just --gpgpu
+        driver = _GpgpuDriver()
+        drivers.append(driver)
 
     if len(drivers) < 1:
         return result
@@ -751,10 +796,9 @@ def gpgpu_install_filter(packages, drivers_str):
     # any package which matches any of those globs will be accepted
     for driver in drivers:
         if driver.flavour:
-            pattern = '%s*%s*' % (driver.vendor, driver.flavour)
+            pattern = '%s*%s' % (driver.vendor, driver.flavour)
         else:
             pattern = '%s*' % (driver.vendor)
-        # print('pattern: %s' % pattern)
         allow.extend(fnmatch.filter(packages, pattern))
         # print(allow)
 
@@ -776,6 +820,35 @@ def gpgpu_install_filter(packages, drivers_str):
     return result
 
 
+def auto_install_filter(packages, drivers_str=''):
+    '''Get packages which are appropriate for automatic installation.
+
+    Return the subset of the given list of packages which are appropriate for
+    automatic installation by the installer. This applies to e. g. the Broadcom
+    Wifi driver (as there is no alternative), but not to the FGLRX proprietary
+    graphics driver (as the free driver works well and FGLRX does not provide
+    KMS).
+    '''
+    # any package which matches any of those globs will be accepted
+    whitelist = ['bcmwl*', 'pvr-omap*', 'virtualbox-guest*', 'nvidia-*',
+                 'open-vm-tools*', 'oem-*-meta']
+
+    # If users specify a driver, use gpgpu_install_filter()
+    if drivers_str:
+        results = gpgpu_install_filter(packages, drivers_str)
+        return results
+
+    allow = []
+    for pattern in whitelist:
+        allow.extend(fnmatch.filter(packages, pattern))
+
+    result = {}
+    for p in allow:
+        if 'recommended' not in packages[p] or packages[p]['recommended']:
+            result[p] = packages[p]
+    return result
+
+
 def detect_plugin_packages(apt_cache=None):
     '''Get driver packages from custom detection plugins.
 
@@ -799,7 +872,11 @@ def detect_plugin_packages(apt_cache=None):
         return packages
 
     if apt_cache is None:
-        apt_cache = apt.Cache()
+        try:
+            apt_cache = apt.Cache()
+        except Exception as ex:
+            logging.error(ex)
+            return {}
 
     for fname in os.listdir(plugindir):
         if not fname.endswith('.py'):
@@ -833,21 +910,63 @@ def detect_plugin_packages(apt_cache=None):
     return packages
 
 
-def _cmp_gfx_alternatives(x, y):
-    '''Compare two graphics driver names in terms of preference.
+def _pkg_support_from_cache(x):
+    '''Look up driver package and return their support level'''
+    if lookup_cache.get(x):
+        return lookup_cache.get(x).get('support')
+    return None
 
-    -updates always sort after non-updates, as we prefer the stable driver and
-    only want to offer -updates when the one from release does not support the
-    card. We never want to recommend -experimental unless it's the only one
-    available, so sort this last.
+
+def _cmp_gfx_alternatives(x, y):
+    '''Compare two graphics driver names in terms of preference. (desktop)
+
+    NFB (New Feature Branch) or the latest is what we recommend on the desktop.
+    -server always sorts after non-server.
+    LTSB (Long Term Support Branch) always sorts before NFB (New Feature Branch).
+    Legacy always sorts before Beta.
     '''
-    if x.endswith('-updates') and not y.endswith('-updates'):
+    if _pkg_support_from_cache(x) == 'Legacy' and _pkg_support_from_cache(y) != 'Legacy':
         return -1
-    if not x.endswith('-updates') and y.endswith('-updates'):
+    if _pkg_support_from_cache(x) != 'Legacy' and _pkg_support_from_cache(y) == 'Legacy':
         return 1
-    if 'experiment' in x and 'experiment' not in y:
+    if _pkg_support_from_cache(x) == 'Beta' and _pkg_support_from_cache(y) != 'Beta':
         return -1
-    if 'experiment' not in x and 'experiment' in y:
+    if _pkg_support_from_cache(x) != 'Beta' and _pkg_support_from_cache(y) == 'Beta':
+        return 1
+    if x.endswith('-server') and not y.endswith('-server'):
+        return -1
+    if not x.endswith('-server') and y.endswith('-server'):
+        return 1
+    if x < y:
+        return -1
+    if x > y:
+        return 1
+    assert x == y
+    return 0
+
+
+def _cmp_gfx_alternatives_gpgpu(x, y):
+    '''Compare two graphics driver names in terms of preference. (server)
+
+    -server always sorts before non-server.
+    LTSB (Long Term Support Branch) always sorts before NFB (New Feature Branch).
+    Legacy always sorts before Beta.
+    '''
+    if x.endswith('-server') and not y.endswith('-server'):
+        return 1
+    if not x.endswith('-server') and y.endswith('-server'):
+        return -1
+    if _pkg_support_from_cache(x) == 'LTSB' and _pkg_support_from_cache(y) != 'LTSB':
+        return 1
+    if _pkg_support_from_cache(x) != 'LTSB' and _pkg_support_from_cache(y) == 'LTSB':
+        return -1
+    if _pkg_support_from_cache(x) == 'Legacy' and _pkg_support_from_cache(y) != 'Legacy':
+        return -1
+    if _pkg_support_from_cache(x) != 'Legacy' and _pkg_support_from_cache(y) == 'Legacy':
+        return 1
+    if _pkg_support_from_cache(x) == 'Beta' and _pkg_support_from_cache(y) != 'Beta':
+        return -1
+    if _pkg_support_from_cache(x) != 'Beta' and _pkg_support_from_cache(y) == 'Beta':
         return 1
     if x < y:
         return -1
@@ -953,6 +1072,12 @@ def get_linux_modules_metapackage(apt_cache, candidate):
         logging.debug('Non NVIDIA linux-modules packages are not supported at this time: %s. Skipping', candidate)
         return metapackage
 
+    pattern = re.compile('nvidia-([0-9]+)')
+    match = pattern.match(candidate)
+    if match:
+        logging.debug('Legacy driver detected: %s. Skipping.', candidate)
+        return metapackage
+
     linux_image_meta = get_linux_image(apt_cache)
     # Check the actual image package, and find the flavour from there
     linux_image = get_linux_image_from_meta(apt_cache, linux_image_meta)
@@ -960,17 +1085,16 @@ def get_linux_modules_metapackage(apt_cache, candidate):
     if linux_image:
         linux_flavour = linux_image.replace('linux-image-', '')
     else:
-        logging.error('No linux-image can be found for %s. Skipping.', candidate)
+        logging.debug('No linux-image can be found for %s. Skipping.', candidate)
         return metapackage
 
-    candidate_flavour = candidate[candidate.rfind('-')+1:]
-
-    try:
-        int(candidate_flavour)
-    except ValueError:
-        logging.error('No flavour can be found in %s. Skipping.', candidate)
+    pattern = re.compile('nvidia-.*-([0-9]+)(.*)')
+    match = pattern.match(candidate)
+    if not match:
+        logging.debug('No flavour can be found in %s. Skipping.', candidate)
         return metapackage
 
+    candidate_flavour = '%s%s' % (match.group(1), match.group(2))
     linux_modules_candidate = 'linux-modules-nvidia-%s-%s' % (candidate_flavour, linux_flavour)
 
     try:
